@@ -339,7 +339,6 @@ renegotiate_data(ConnectionHandler) ->
 	  ssh_params                            :: #ssh{}
 						 | undefined,
 	  socket                                :: inet:socket(),
-	  sender :: pid() | undefined,
 	  decrypted_data_buffer     = <<>>      :: binary(),
 	  encrypted_data_buffer     = <<>>      :: binary(),
 	  undecrypted_packet_length             :: undefined | non_neg_integer(),
@@ -368,10 +367,9 @@ init_connection_handler(Role, Socket, Opts) ->
 	{Protocol, Callback, CloseTag} =
 	    proplists:get_value(transport, Opts, ?DefaultTransport),
 	S0#data{ssh_params = init_ssh_record(Role, Socket, Opts),
-		sender = spawn_link(fun() -> nonblocking_sender(Socket, Callback) end),
-		transport_protocol = Protocol,
-		transport_cb = Callback,
-		transport_close_tag = CloseTag
+		 transport_protocol = Protocol,
+		 transport_cb = Callback,
+		 transport_close_tag = CloseTag
 		}
     of
 	S ->
@@ -547,6 +545,7 @@ handle_event(_, {info_line,_Line}, {hello,Role}, D) ->
     case Role of
 	client ->
 	    %% The server may send info lines to the client before the version_exchange
+	    %% RFC4253/4.2
 	    inet:setopts(D#data.socket, [{active, once}]),
 	    keep_state_and_data;
 	server ->
@@ -672,8 +671,9 @@ handle_event(_, #ssh_msg_newkeys{} = Msg, {new_keys,Role,init}, D) ->
     {next_state, {service_request,Role}, D#data{ssh_params=Ssh}};
 
 %% Subsequent key exchange rounds (renegotiation):
-handle_event(_, #ssh_msg_newkeys{}, {new_keys,Role,renegotiate}, D) ->
-    {next_state, {connected,Role}, D};
+handle_event(_, #ssh_msg_newkeys{} = Msg, {new_keys,Role,renegotiate}, D) ->
+    {ok, Ssh} = ssh_transport:handle_new_keys(Msg, D#data.ssh_params),
+    {next_state, {connected,Role}, D#data{ssh_params=Ssh}};
 
 %%% ######## {service_request, client|server}
 
@@ -928,6 +928,7 @@ handle_event(internal, Msg=#ssh_msg_channel_request{},           StateName, D) -
     handle_connection_msg(Msg, StateName, D);
 
 handle_event(internal, Msg=#ssh_msg_channel_success{},           StateName, D) ->
+    update_inet_buffers(D#data.socket),
     handle_connection_msg(Msg, StateName, D);
 
 handle_event(internal, Msg=#ssh_msg_channel_failure{},           StateName, D) ->
@@ -1007,6 +1008,7 @@ handle_event(cast, {reply_request,success,ChannelId}, {connected,_}, D) ->
     case ssh_channel:cache_lookup(cache(D), ChannelId) of
 	#channel{remote_id = RemoteId} ->
 	    Msg = ssh_connection:channel_success_msg(RemoteId),
+	    update_inet_buffers(D#data.socket),
 	    {keep_state, send_msg(Msg,D)};
 
 	undefined ->
@@ -1194,12 +1196,12 @@ handle_event(info, {Proto, Sock, NewData}, StateName, D0 = #data{socket = Sock,
 		ssh_message:decode(set_kex_overload_prefix(DecryptedBytes,D))
 	    of
 		Msg = #ssh_msg_kexinit{} ->
-		    {keep_state, D, [{next_event, internal, {Msg,DecryptedBytes}},
-				     {next_event, internal, prepare_next_packet}
+		    {keep_state, D, [{next_event, internal, prepare_next_packet},
+				     {next_event, internal, {Msg,DecryptedBytes}}
 				    ]};
 		Msg ->
-		    {keep_state, D, [{next_event, internal, Msg},
-				     {next_event, internal, prepare_next_packet}
+		    {keep_state, D, [{next_event, internal, prepare_next_packet},
+				     {next_event, internal, Msg}
 				    ]}
 	    catch
 		_C:_E  ->
@@ -1447,15 +1449,18 @@ start_the_connection_child(UserPid, Role, Socket, Options) ->
 %% Stopping
 -type finalize_termination_result() :: ok .
 
-finalize_termination(_StateName, D) ->
-    case D#data.connection_state of
+finalize_termination(_StateName, #data{transport_cb = Transport,
+				       connection_state = Connection,
+				       socket = Socket}) ->
+    case Connection of
 	#connection{system_supervisor = SysSup,
 		    sub_system_supervisor = SubSysSup} when is_pid(SubSysSup) ->
 	    ssh_system_sup:stop_subsystem(SysSup, SubSysSup);
 	_ ->
 	    do_nothing
     end,
-    close_transport(D).
+    (catch Transport:close(Socket)),
+    ok.
 
 %%--------------------------------------------------------------------
 %% "Invert" the Role
@@ -1510,34 +1515,10 @@ send_msg(Msg, State=#data{ssh_params=Ssh0}) when is_tuple(Msg) ->
     send_bytes(Bytes, State),
     State#data{ssh_params=Ssh}.
 
-send_bytes(Bytes, #data{sender = Sender}) ->
-    Sender ! {send,Bytes},
+send_bytes(Bytes, #data{socket = Socket, transport_cb = Transport}) ->
+    _ = Transport:send(Socket, Bytes),
     ok.
 
-close_transport(D) ->
-    D#data.sender ! close,
-    ok.
-
-
-nonblocking_sender(Socket, Callback) ->
-    receive
-	{send, Bytes} ->
-	    case Callback:send(Socket, Bytes) of
-	    	ok ->
-	    	    nonblocking_sender(Socket, Callback);
-	    	E = {error,_} ->
-	    	    exit({shutdown,E})
-	    end;
-	
-	close ->
-	    case Callback:close(Socket) of
-		ok ->
-		    ok;
-		E = {error,_} ->
-	    	    exit({shutdown,E})
-	    end
-    end.
-
 handle_version({2, 0} = NumVsn, StrVsn, Ssh0) ->
     Ssh = counterpart_versions(NumVsn, StrVsn, Ssh0),
     {ok, Ssh};
@@ -1759,6 +1740,11 @@ send_replies(Repls, State) ->
 		Repls).
 
 get_repl({connection_reply,Msg}, {CallRepls,S}) ->
+    if is_record(Msg, ssh_msg_channel_success) ->
+	    update_inet_buffers(S#data.socket);
+       true ->
+	    ok
+    end,
     {CallRepls, send_msg(Msg,S)};
 get_repl({channel_data,undefined,_Data}, Acc) ->
     Acc;
@@ -1947,3 +1933,13 @@ handshake(Pid, Ref, Timeout) ->
 	    {error, timeout}
     end.
 
+update_inet_buffers(Socket) ->
+    {ok, BufSzs0} = inet:getopts(Socket, [sndbuf,recbuf]),
+    MinVal = 655360,
+    case
+	[{Tag,MinVal} || {Tag,Val} <- BufSzs0,
+			 Val < MinVal]
+    of
+	[] -> ok;
+	NewOpts -> inet:setopts(Socket, NewOpts)
+    end.
